@@ -182,8 +182,8 @@ namespace eval ::pki {
 		2.5.29.18                      id-ce-issuerAltName
 		2.5.29.19                      id-ce-basicConstraints
 		2.5.29.20                      id-ce-cRLNumber
+		2.5.29.31                      id-ce-cRLDistributionPoints
 		2.5.29.32                      id-ce-certificatePolicies
-		2.5.29.33                      id-ce-cRLDistributionPoints
 		2.5.29.35                      id-ce-authorityKeyIdentifier
 	}
 
@@ -268,7 +268,7 @@ proc ::pki::_oid_name_to_number {name} {
 		}
 	}
 
-	return -code error
+	return -code error "Unable to convert OID $name to an OID value"
 }
 
 proc ::pki::rsa::_encrypt_num {input exponent mod} {
@@ -1627,8 +1627,24 @@ proc ::pki::x509::create_cert {signreqlist cakeylist serial_number notBefore not
 	lappend certlist [::asn::asnBigInteger [math::bignum::fromstr $serial_number]]
 
 	## Insert data algorithm
+	switch -glob -- $algo {
+		"*With*Encryption" {
+			# Already fully qualified, add nothing
+			regexp {^.*With(.*)Encryption$} $algo -> type
+			regexp {^(.*)With(.*)Encryption$} $algo -> hashingAlgorithm type
+
+			set hashingAlgorithm [string tolower $hashingAlgorithm]
+			set type [string tolower $type]
+		}
+		default {
+			set hashingAlgorithm $algo
+
+			set algo "${algo}With${type}Encryption"
+		}
+	}
+
 	lappend certlist [::asn::asnSequence \
-		[::asn::asnObjectIdentifier [::pki::_oid_name_to_number "${algo}With${type}Encryption"]] \
+		[::asn::asnObjectIdentifier [::pki::_oid_name_to_number "${algo}"]] \
 		[::asn::asnNull] \
 	]
 
@@ -1692,10 +1708,10 @@ proc ::pki::x509::create_cert {signreqlist cakeylist serial_number notBefore not
 					foreach {altnametype altnamevalue} [lrange $extvalue 1 end] {
 						switch -- [string tolower $altnametype] {
 							"rfc822name" {
-								lappend altnames [::asn::asnChoice 1 [::asn::asnIA5String $altnamevalue]]
+								lappend altnames [::asn::asnChoice 1 $altnamevalue]
 							}
 							"dnsname" {
-								lappend altnames [::asn::asnChoice 2 [::asn::asnIA5String $altnamevalue]]
+								lappend altnames [::asn::asnChoice 2 $altnamevalue]
 							}
 							default {
 								return -code error "Unknown subjectAltName type: $altnametype"
@@ -1705,6 +1721,86 @@ proc ::pki::x509::create_cert {signreqlist cakeylist serial_number notBefore not
 
 					set extvalue [::asn::asnSequence {*}$altnames]
 				}
+				"id-ce-cRLDistributionPoints" {
+					set critical [lindex $extvalue 0]
+
+					set crlDistributionPoint_objects [list distributionPoint reasons cRLIssuer]
+					set crlDistributionPointsASN [list]
+
+					foreach crlDistributionPoint_dict [lrange $extvalue 1 end] {
+						set crlDistributionPoint_objectASN [list]
+						set crlSequenceIdx -1
+
+						foreach crlDistributionPoint_objectName $crlDistributionPoint_objects {
+							unset -nocomplain crlDistributionPointASN
+							incr crlSequenceIdx
+
+							if {![dict exists $crlDistributionPoint_dict $crlDistributionPoint_objectName]} {
+								continue
+							}
+
+							set crlDistributionPoint_object [dict get $crlDistributionPoint_dict $crlDistributionPoint_objectName]
+
+							switch -- $crlDistributionPoint_objectName {
+								"distributionPoint" {
+									foreach {crlDistributionPointNameType crlDistributionPointName_dict} $crlDistributionPoint_object {
+										switch -- $crlDistributionPointNameType {
+											"name" {
+												unset -nocomplain crlDistributionPointName
+												array set crlDistributionPointName $crlDistributionPointName_dict
+
+												switch -- $crlDistributionPointName(type) {
+													"url" {
+														set crlDistributionPointNameASN [::asn::asnChoice 6 $crlDistributionPointName(value)]
+													}
+													default {
+														return -code error "Unsupported crlDistributionPointName choice: $crlDistributionPointName(type)"
+													}
+												}
+												set crlDistributionPointASN [::asn::asnContextConstr 0 $crlDistributionPointNameASN]
+											}
+											default {
+												return -code error "Unsupported crlDistributionPointNameType: $crlDistributionPointNameType"
+											}
+										}
+									}
+								}
+								default {
+									return -code error "Unsupported crlDistributionPoint option: $crlDistributionPoint_objectName"
+								}
+							}
+
+							lappend crlDistributionPoint_objectASN [::asn::asnContextConstr $crlSequenceIdx $crlDistributionPointASN]
+						}
+						lappend crlDistributionPointsASN [::asn::asnSequenceFromList $crlDistributionPoint_objectASN]
+					}
+
+					set extvalue [::asn::asnSequenceFromList $crlDistributionPointsASN]
+				}
+				"id-ce-keyUsage" {
+					set critical [lindex $extvalue 0]
+					set extvalue [string tolower [lrange $extvalue 1 end]]
+
+					set keyUsages [string tolower [list digitalSignature nonRepudiation keyEncipherment dataEncipherment keyAgreement keyCertSign cRLSign encipherOnly decipherOnly]]
+
+					foreach keyUsage $extvalue {
+						if {$keyUsage ni $keyUsages} {
+							return -code error "Invalid key usage: $keyUsage"
+						}
+					}
+
+					set keyUsageValue ""
+					foreach keyUsage $keyUsages {
+						if {$keyUsage in $extvalue} {
+							set keyUsageResult 1
+						} else {
+							set keyUsageResult 0
+						}
+						append keyUsageValue $keyUsageResult
+					}
+
+					set extvalue [::asn::asnBitString $keyUsageValue]
+				}
 				default {
 					return -code error "Unknown extension: $extension"
 				}
@@ -1724,13 +1820,13 @@ proc ::pki::x509::create_cert {signreqlist cakeylist serial_number notBefore not
 	set cert [::asn::asnSequenceFromList $certlist]
 
 	# Sign certificate request using CA
-	set signature [::pki::sign $cert $cakeylist $algo]
+	set signature [::pki::sign $cert $cakeylist $hashingAlgorithm]
 	binary scan $signature B* signature_bitstring
 
 	set cert [::asn::asnSequence \
 		$cert \
 		[::asn::asnSequence \
-			[::asn::asnObjectIdentifier [::pki::_oid_name_to_number "${algo}With${type}Encryption"]] \
+			[::asn::asnObjectIdentifier [::pki::_oid_name_to_number "${algo}"]] \
 			[::asn::asnNull] \
 		] \
 		[::asn::asnBitString $signature_bitstring] \
@@ -1942,7 +2038,16 @@ proc ::pki::rsa::generate {bitlength {exponent 0x10001}} {
 		}
 
 		if {$plen >= $componentbitlen && $qlen >= $componentbitlen} {
-			break
+			set n [expr {$p * $q}]
+			set nlen [::pki::_bits $n]
+
+			unset n
+
+			if {$nlen == $bitlength} {
+				unset nlen
+
+				break
+                        }
 		}
 
 		set x [::pki::_random]
@@ -2024,4 +2129,4 @@ proc ::pki::rsa::generate {bitlength {exponent 0x10001}} {
 # # ## ### ##### ######## #############
 ## Ready
 
-package provide pki 0.7
+package provide pki 0.8
